/***********************************************

Author contact information: 
    Michael J. Black
    Department of Computer Science
    Brown University 
    P.O. Box 1910
    Providence, RI 02912 

    http://www.cs.brown.edu/people/black/
    email: black@cs.brown.edu

    401-863-7637 (voice) 
    401-863-7657 (fax) 

---------------------------------------------------------------------
Copyright 1993 Michael J. Black

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any non-commercial purpose is hereby granted without
fee, provided that the above copyright notice appear in all copies and
that both that copyright notice and this permission notice appear in
supporting documentation, and that the name of the author not be used
in advertising or publicity pertaining to distribution of the software
without specific, written prior permission.

THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
ANY PARTICULAR PURPOSE.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

For commercial use and licensing information contact the author
at the address above
---------------------------------------------------------------------

************************************************/

# include       <stdio.h>
# include       <math.h>
# include       <string.h>

#include "robust.h"
#include "planar.h"

main(argc, argv)
 
     int argc;
     char *argv[];

{
    float *image1, *image2, *Ix, *Iy, *It, *outlier_img, *weights, *lap1, *lap2, *dummy;
    float *warped, *used_pts, *mean;
    float a0=0.0, a1=0.0, a2=0.0, a3=0.0, a4=0.0, a5=0.0, a6=0.0, a7=0.0, u=0.0, v=0.0;
    float sigma=1.0, sigma_init=1.0, omega=1.995;     
    int size, nx, ny, i, j, fdCurr, skip1=0, skip2=0;
    int levels=0, iters=1, filter=0, affine=0, planar=0, multiple=0;
    char fnCurr[MAXLEN];
    char *inpath1, *inpath2, *outpath, *flag;
    float *load_image(), rate=1.0, outlier_to_sigma();

    inpath1 = *++argv;
    inpath2 = *++argv;
    outpath = *++argv;
/*
    strcpy(fnCurr,inpath1);
    image1 = load_image(fnCurr, image1); 
    strcpy(fnCurr,inpath2);
    image2 = load_image(fnCurr, image2); 
*/
    argc = argc - 3;
    while (--argc > 0 && (*++argv)[0] == '-') {
      flag = ++argv[0];
      if (strcmp(flag, "nx") == 0) {
	nx = atoi(*++argv);
	fprintf(stderr, "nx: %d\n", nx);
      }
      else if (strcmp(flag, "ny") == 0) {
	ny = atoi(*++argv);
	fprintf(stderr, "ny: %d\n", ny);
      }
      else if (strcmp(flag, "iters") == 0) {
	iters = atoi(*++argv);
	fprintf(stderr, "iters: %d\n", iters);
      }
      else if (strcmp(flag, "s") == 0) {
	sigma = atof(*++argv);
	fprintf(stderr, "sigma: %f\n", sigma);
      }
      else if (strcmp(flag, "S") == 0) {
	sigma_init = atof(*++argv);
	fprintf(stderr, "intial sigma: %f\n", sigma_init);
      }
      else if (strcmp(flag, "w") == 0) {
	omega = atof(*++argv);
	fprintf(stderr, "omega: %f\n", omega);
      }
      else if (strcmp(flag, "u") == 0) {
	a0 = atof(*++argv);
	u = a0;
	fprintf(stderr, "u: %f\n", a0);
      }
      else if (strcmp(flag, "v") == 0) {
	a3 = atof(*++argv);
	v = a3;
	fprintf(stderr, "v: %f\n", a3);
      }
      else if (strcmp(flag, "f") == 0) {
	filter = atoi(*++argv);
	fprintf(stderr, "filter: %d\n", filter);
      }
      else if (strcmp(flag, "a") == 0) {
	affine = atoi(*++argv);
	fprintf(stderr, "affine: %d\n", affine);
      }
      else if (strcmp(flag, "p") == 0) {
	planar = atoi(*++argv);
	fprintf(stderr, "planar: %d\n", planar);
      }
      else if (strcmp(flag, "m") == 0) {
	multiple = atoi(*++argv);
	fprintf(stderr, "multiple: %d\n", multiple);
      }
      else if (strcmp(flag, "l") == 0) {
	levels = atoi(*++argv);
	fprintf(stderr, "levels: %d\n", levels);
      }
      else if (strcmp(flag, "r") == 0) {
	rate = atof(*++argv);
	fprintf(stderr, "rate: %f\n", rate);
      }
      else if (strcmp(flag, "skip1") == 0) {
	skip1 = atoi(*++argv);
	fprintf(stderr, "skip1: %d\n", skip1);
      }
      else if (strcmp(flag, "skip2") == 0) {
	skip2 = atoi(*++argv);
	fprintf(stderr, "skip2: %d\n", skip2);
      }
      argc = argc -1;
    }

    sigma = outlier_to_sigma(sigma);
    sigma_init = outlier_to_sigma(sigma_init);

    size = nx * ny * sizeof( float );
    if ((lap1 = (float *) malloc((size_t) size)) == NULL)
	error(" Unable to allocate memory for ", "lap1");
    if ((lap2 = (float *) malloc((size_t) size)) == NULL)
	error(" Unable to allocate memory for ", "lap2");
    if ((Ix = (float *) malloc((size_t) size)) == NULL)
	error(" Unable to allocate memory for ", "Ix");
    if ((Iy = (float *) malloc((size_t) size)) == NULL)
	error(" Unable to allocate memory for ", "Iy");
    if ((It = (float *) malloc((size_t) size)) == NULL)
	error(" Unable to allocate memory for ", "It");
    if ((outlier_img = (float *) malloc((size_t) size)) == NULL)
	error(" Unable to allocate memory for ", "outlier_img");
    if ((weights = (float *) malloc((size_t) size)) == NULL)
	error(" Unable to allocate memory for ", "weights");
    if ((warped = (float *) malloc((size_t) size)) == NULL)
	error(" Unable to allocate memory for ", "warped");
    if ((mean = (float *) malloc((size_t) size)) == NULL)
	error(" Unable to allocate memory for ", "mean");
    if ((used_pts = (float *) malloc((size_t) size)) == NULL)
	error(" Unable to allocate memory for ", "used_pts");
    if ((dummy = (float *) malloc((size_t) size)) == NULL)
	error(" Unable to allocate memory for ", "dummy");

    if ((image1 = (float *) malloc((size_t) size)) == NULL)
      error(" Unable to allocate memory for ", "image1");
    if ((image2 = (float *) malloc((size_t) size)) == NULL)
      error(" Unable to allocate memory for ", "image2");

    strcpy(fnCurr,inpath1);
    if ( (fdCurr = open(fnCurr,0)) == -1 ) 
      error("Can't open input file ",fnCurr);
    read_image_strip_bytes(image1, fdCurr, fnCurr, nx, ny, skip1);
    /* read_image(image1, fdCurr, fnCurr, nx, ny); */
    close(fdCurr);

    strcpy(fnCurr,inpath2);
    if ( (fdCurr = open(fnCurr,0)) == -1 ) 
      error("Can't open input file ",fnCurr);
    read_image_strip_bytes(image2, fdCurr, fnCurr, nx, ny, skip2);
    /* read_image(image2, fdCurr, fnCurr, nx, ny); */
    close(fdCurr);

/*
    strcpy(fnCurr,outpath);
    strcat(fnCurr, "-image1");
    save_pgm(fnCurr, image1, nx, ny);

    strcpy(fnCurr,outpath);
    strcat(fnCurr, "-image2");
    save_pgm(fnCurr, image2, nx, ny);
*/
    for(i=0;i<ny;i++){
      for(j=0;j<nx;j++){
	outlier_img[(i*nx)+j] = 1.0;
	weights[(i*nx)+j] = 0.0;
      }
    }
    /* fprintf(stderr, "regression\n"); */
    pyramid_regress(image1, image2, levels, &a0, &a1, &a2, &a3, &a4, &a5,
		    &a6, &a7, Ix, Iy, It, nx, ny, 
		    iters, sigma, sigma_init, omega, outlier_img, weights,
		    filter, affine, planar, rate, 0, 0, nx, ny);


    planar_warp_image(image2, warped, a0, a1, a2, a3, a4, a5, a6, a7, 
		      nx, ny);
    planar_warp_outliers(a0, a1, a2, a3, a4, a5, a6, a7, 
			 nx, ny,  outlier_img, 1.0);

    copy_image(used_pts, outlier_img, nx, ny);
    strcpy(fnCurr,outpath);
    strcat(fnCurr, "-outliers1");
    save_pgm(fnCurr, outlier_img, nx, ny);
    invert_image(outlier_img, nx, ny);

    strcpy(fnCurr,outpath);
    strcat(fnCurr, "-weights1");
    save_pgm(fnCurr, weights, nx, ny);
/*
    outliers(Ix, Iy, It, a0, a1, a2, a3, a4, a5, nx, ny, sigma, outlier_img);
    save_pgm("outliers", outlier_img, nx, ny);
*/
    strcpy(fnCurr,outpath);
    strcat(fnCurr, "-stable1");
    save_pgm(fnCurr, warped, nx, ny);
    average_images(warped, image1, mean, nx, ny);
    strcpy(fnCurr,outpath);
    strcat(fnCurr, "-mean1");
    save_pgm(fnCurr, mean, nx, ny);
    subtract_image(warped, image1, nx, ny);
    strcpy(fnCurr,outpath);
    strcat(fnCurr, "-warped1");
    save_pgm(fnCurr, warped, nx, ny);
/*    
    save_pgm("im1", image1, nx, ny);
    save_pgm("im2", image2, nx, ny);
*/
    /* compensate for the dominant motion */
    planar_warp_image(image2, warped, a0, a1, a2, a3, a4, a5, a6, a7, 
		      nx, ny);
    copy_image(image2, warped, nx, ny);

    /* compute relative affine motions wrt dominant */
    if (multiple>=1) {
      fprintf(stderr, "\n LEVEL 1\n");
      a0=u; a1=0.0; a2=0.0;
      a3=v; a4=0.0; a5=0.0;
      a6=0.0; a7=0.0;
      pyramid_regress(image1, image2, levels, &a0, &a1, &a2, &a3, &a4, &a5,
		      &a6, &a7, Ix, Iy, It, nx, ny, 
		      iters, (sigma/1.0), sigma_init, omega, 
		      outlier_img, weights, filter, affine, planar, rate, 0, 0, nx, ny);
      strcpy(fnCurr,outpath);
      strcat(fnCurr, "-outliers2");
      subtract_image(outlier_img, used_pts, nx, ny);
      save_pgm(fnCurr, outlier_img, nx, ny);

      planar_warp_image(image2, warped, a0, a1, a2, a3, a4, a5, a6, a7, 
			nx, ny);
      strcpy(fnCurr,outpath);
      strcat(fnCurr, "-stable2");
      save_pgm(fnCurr, warped, nx, ny);
      average_images(warped, image1, mean, nx, ny);
      strcpy(fnCurr,outpath);
      strcat(fnCurr, "-mean2");
      save_pgm(fnCurr, mean, nx, ny);
      subtract_image(warped, image1, nx, ny);
      strcpy(fnCurr,outpath);
      strcat(fnCurr, "-warped2");
      save_pgm(fnCurr, warped, nx, ny);
      strcpy(fnCurr,outpath);
      strcat(fnCurr, "-weights2");
      save_pgm(fnCurr, weights, nx, ny);
    }
   
    if (multiple>=2) {
      fprintf(stderr, "\n LEVEL 2\n");
      
      add_image(used_pts, outlier_img, nx, ny);
      copy_image(outlier_img, used_pts, nx, ny);
      invert_image(outlier_img, nx, ny);
      strcpy(fnCurr,outpath);
      strcat(fnCurr, "-outliers3-in");
      save_pgm(fnCurr, outlier_img, nx, ny);

      a0=u; a1=0.0; a2=0.0;
      a3=v; a4=0.0; a5=0.0;
      a6=0.0; a7=0.0;
      pyramid_regress(image1, image2, levels, &a0, &a1, &a2, &a3, &a4, &a5,
		      &a6, &a7, Ix, Iy, It, nx, ny, 
		      iters, (sigma/1.0), sigma_init, omega, outlier_img, weights,
		      filter, affine, planar, rate, 0, 0, nx, ny);
      strcpy(fnCurr,outpath);
      strcat(fnCurr, "-outliers3");
      subtract_image(outlier_img, used_pts, nx, ny);
      save_pgm(fnCurr, outlier_img, nx, ny);

      planar_warp_image(image2, warped, a0, a1, a2, a3, a4, a5, a6, a7, 
			nx, ny);
      strcpy(fnCurr,outpath);
      strcat(fnCurr, "-stable3");
      save_pgm(fnCurr, warped, nx, ny);
      average_images(warped, image1, mean, nx, ny);
      strcpy(fnCurr,outpath);
      strcat(fnCurr, "-mean3");
      save_pgm(fnCurr, mean, nx, ny);
      subtract_image(warped, image1, nx, ny);
      strcpy(fnCurr,outpath);
      strcat(fnCurr, "-warped3");
      save_pgm(fnCurr, warped, nx, ny);
      strcpy(fnCurr,outpath);
      strcat(fnCurr, "-weights3");
      save_pgm(fnCurr, weights, nx, ny);
    }
   
    fprintf(stderr, "done\n");
  }




    /* for testing:
    affine_warp_image(image1, image2, -1.0, 0.0001, 0.002, 
		      0.0, 0.00007, 0.0005, nx, ny);
    */
    /*
    if (filter == 1) {
      fprintf(stderr, "filtered\n");
      cmvu_laplacian(image1, lap1, nx, ny);
      cmvu_laplacian(image2, lap2, nx, ny);
    }
    else {
      copy_image(lap1, image1, nx, ny);
      copy_image(lap2, image2, nx, ny);
    }
    */

/*
    robust_regression(lap1, lap2, &a0, &a3, Ix, Iy, It, nx, ny, 
		      iters, sigma, omega, outlier_img);

    if (affine==1) {
      regress_affine(lap1, lap2, a0, &a1, &a2, a3, &a4, &a5,
		     Ix, Iy, It, nx, ny, 
		     iters, sigma, omega, outlier_img);
    }
*/

/*

      robust_regression(lap1, lap2, &a0, &a3, Ix, Iy, It, nx, ny, 
			iters, sigma, omega, outlier_img);
      if (affine==1) {
	regress_affine(lap1, lap2, a0, &a1, &a2, a3, &a4, &a5,
		       Ix, Iy, It, nx, ny, 
		       iters, sigma, omega, outlier_img);
      }

      affine_warp_image(image1, warped, a0, a1, a2, a3, a4, a5, nx, ny);
      subtract_image(warped, image2, nx, ny);
      save_pgm("warped2", warped, nx, ny);
    }
*/

/*
gcc new-image-regress.c gm2.c planar-warp.c ~/IGNC/src/New/derivatives.c ~/IGNC/src/New/image-io.c ~/IGNC/src/New/utils.c ~/IGNC/src/New/pyramid.c -o planar -lm

gcc new-image-regress.c planar-regression.c gm2.c Quadratic/bicubic-interp.c ~/IGNC/src/New/eero-filter.c ~/IGNC/src/New/derivatives.c ~/IGNC/src/New/image-io.c ~/IGNC/src/New/utils.c ~/IGNC/src/New/pyramid.c -o planar2 -lm

regress /usr/people/black/data/rdot-cvpr92/rdot1 /usr/people/black/data/rdot-cvpr92/rdot2 -iters 1 -nx 128 -ny 128

regress /usr/people/black/data/pepsi/pepsi-0 /usr/people/black/data/pepsi/pepsi-2 -iters 50 -s 1.5 -nx 64 -ny 64 -u -0.0 -f 0

regress ../data/escher0 ../data/escher1 -iters 50 -s 1.5 -nx 360 -ny 400 -u -0.0 -f 0

regress ../data/small0 ../data/small1 -iters 50 -s 1.5 -nx 180 -ny 200 -u -1.0 -f 0
*/
