/***********************************************

Author contact information: 
    Michael J. Black
    Department of Computer Science
    Brown University 
    P.O. Box 1910
    Providence, RI 02912 

    http://www.cs.brown.edu/people/black/
    email: black@cs.brown.edu

    401-863-7637 (voice) 
    401-863-7657 (fax) 

---------------------------------------------------------------------
Copyright 1993, 2002 Michael J. Black

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any non-commercial purpose is hereby granted without
fee, provided that the above copyright notice appear in all copies and
that both that copyright notice and this permission notice appear in
supporting documentation, and that the name of the author not be used
in advertising or publicity pertaining to distribution of the software
without specific, written prior permission.

THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
ANY PARTICULAR PURPOSE.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

For commercial use and licensing information contact the author
at the address above
---------------------------------------------------------------------

************************************************/

# include       <stdio.h>
# include       <math.h>
# include       <string.h>

#include "robust.h"
#include "planar.h"

pyramid_regress(im1, im2, level,
		a0, a1, a2, a3, a4, a5, a6, a7, Ix, Iy, It, nx, ny, 
		iters, sigma, sigma_init, omega, omask, weights, filter, 
		affine, planar, rate)
     float *im1, *im2, *a0, *a1, *a2, *a3, *a4, *a5, *a6, *a7;
     float *Ix, *Iy, *It, sigma, sigma_init, omega, *omask, *weights;
     int iters, nx, ny, level, filter, affine, planar;
     float rate;
{
  float *p_a0, *p_a3;
  float *p_im1, *p_im2, *p_out_im, *warp, *lap1, *lap2;
  int size, small_size;
  float da0=0.0, da1=0.0, da2=0.0, da3=0.0, da4=0.0, da5=0.0, da6=0.0, da7=0.0;
  float regress_affine(), min_sigma;
  double floor();

  fprintf(stderr," nx %d, ny %d\n", nx, ny); 
  fprintf(stderr, "initial values: %f %f %f %f %f %f %f %f\n", 
	  *a0, *a1, *a2, *a3, *a4, *a5, *a6, *a7);
/*
  fprintf(stderr, "Region: (%d, %d, %d, %d)\n",
	  x, y, x+dx, y+dy);
*/
/*
  save_image("im1", im1, nx, ny);
  save_image("im2", im2, nx, ny);
*/
  size = nx * ny * sizeof( float );
  small_size = (int) floor(nx/2.0) * (int)floor(ny/2.0) * sizeof( float );
  /* fprintf(stderr, "small_size: %d\n", small_size); */
  if ((p_im1 = (float *) malloc((size_t) small_size)) == NULL)
    error(" Unable to allocate memory for ", "p_im1");
  if ((p_im2 = (float *) malloc((size_t) small_size)) == NULL)
    error(" Unable to allocate memory for ", "p_im2");
  if ((p_out_im = (float *) malloc((size_t) small_size)) == NULL)
    error(" Unable to allocate memory for ", "p_out_im");
  if ((warp = (float *) malloc((size_t) size)) == NULL)
    error(" Unable to allocate memory for ", "warp");
  if ((lap1 = (float *) malloc((size_t) size)) == NULL)
    error(" Unable to allocate memory for ", "lap1");
  if ((lap2 = (float *) malloc((size_t) size)) == NULL)
    error(" Unable to allocate memory for ", "lap2");

  if (level <= 0) {
    planar_warp_image(im2, warp, *a0, *a1, *a2, *a3, *a4, *a5, *a6, *a7, 
		      nx, ny); 
    if (filter == 1) {
      cmvu_laplacian(warp, lap2, nx, ny);
      cmvu_laplacian(im1, lap1, nx, ny);
      /*
      cmvu_laplacian(warp, lap2, nx, ny);
      cmvu_laplacian(im1, lap1, nx, ny);
      */
    }
    else {
      copy_image(lap2, warp, nx, ny);
      copy_image(lap1, im1, nx, ny);
    }
    horn_dx(lap1, lap2, Ix, nx, ny);
    horn_dy(lap1, lap2, Iy, nx, ny);
    horn_dt(lap1, lap2, It, nx, ny);

    min_sigma = regress_affine(lap1, lap2, &da0, &da1, &da2, &da3, &da4, &da5,
		   &da6, &da7, Ix, Iy, It, nx, ny, 
		   iters, sigma, sigma_init, omega, omask, rate, affine, planar);
    *a0 += da0; *a3 += da3; 
    *a1 += da1; *a2 += da2;
    *a4 += da4; *a5 += da5;
    *a6 += da6; *a7 += da7;
  }
  else {
    /* recurse */
    reduce_image(im1, p_im1, nx, ny);
    fprintf(stderr, "reduced\n");
    reduce_image(im2, p_im2, nx, ny);
    fprintf(stderr, "reduced\n");
    reduce_image(omask, p_out_im, nx, ny);
    fprintf(stderr, "reduced\n");
    *a0 = (*a0)/2.0;  
    *a3 = (*a3)/2.0;
/*    
    fprintf(stderr, "%d,  %d\n", nx, ny);
    fprintf(stderr, "%d,  %d\n", (int) floor(nx / 2.0), (int)floor(ny / 2.0));
*/
    fprintf(stderr, "recursing...\n");
    pyramid_regress(p_im1, p_im2, (level - 1),
		    a0, a1, a2, a3, a4, a5, a6, a7,
		    Ix, Iy, It, (int)floor(nx / 2.0), (int)floor(ny / 2.0), 
		    iters, sigma, sigma_init, omega, 
		    p_out_im, weights, filter, affine, planar, rate);
    *a0 = *a0 * 2.0; 
    *a3 = *a3 * 2.0; 

    fprintf(stderr, "returned %f %f %f %f %f %f %f %f\n", 
	  *a0, *a1, *a2, *a3, *a4, *a5, *a6, *a7);
    
    /* update */
    planar_warp_image(im2, warp, *a0, *a1, *a2, *a3, *a4, *a5, *a6, *a7, 
		      nx, ny);
    if (filter == 1) {
      cmvu_laplacian(warp, lap2, nx, ny);
      cmvu_laplacian(im1, lap1, nx, ny);
      /*
      cmvu_laplacian(warp, lap2, nx, ny);
      cmvu_laplacian(im1, lap1, nx, ny);
      */
    }
    else {
      copy_image(lap2, warp, nx, ny);
      copy_image(lap1, im1, nx, ny);
    }
    horn_dx(lap1, lap2, Ix, nx, ny);
    horn_dy(lap1, lap2, Iy, nx, ny);
    horn_dt(lap1, lap2, It, nx, ny);
    
    da0 = 0.0; da3=0.0;
    min_sigma = regress_affine(lap1, lap2, &da0, &da1, &da2, &da3, &da4, &da5,
		   &da6, &da7, Ix, Iy, It, nx, ny, 
		   iters, sigma, sigma_init, omega, omask, rate, affine, planar);
    *a0 += da0; *a3 += da3; 
    *a1 += da1; *a2 += da2;
    *a4 += da4; *a5 += da5;
    *a6 += da6; *a7 += da7;

    fprintf(stderr, "updated %f %f %f %f %f %f %f %f\n", 
	    *a0, *a1, *a2, *a3, *a4, *a5, *a6, *a7);
  }

  /* save_image("/tmp/black/tmp-outliers", omask, nx, ny);  */
  
  horn_dx(lap1, lap2, Ix, nx, ny);
  horn_dy(lap1, lap2, Iy, nx, ny);
  horn_dt(lap1, lap2, It, nx, ny);
  good_points(Ix, Iy, It, da0, da1, da2, da3, da4, da5, da6, da7, nx, ny, min_sigma, omask);

  robust_to_weights(Ix, Iy, It, da0, da1, da2, da3, da4, da5, da6, da7, nx, ny, min_sigma, weights);
    
  free(p_im1);
  free(p_im2);
  free(p_out_im);
  free(warp);
  free(lap1);
  free(lap2);
}



float affine_regress_1_iter(Ix, Iy, It, a0, a1, a2, a3, a4, a5, a6, a7,
			    nx, ny, sigma, omega, omask, affine, planar)

     float *Ix, *Iy, *It, *a0, *a1, *a2, *a3, *a4, *a5, *a6, *a7, sigma, omega;
     float *omask;
     int nx, ny, affine, planar;
{
  int index, i, j, x, y, x_cent, y_cent, num_outliers=0;
  float err, delta, max_delta=0.0, top, scale, gr;
  float psi(), dpsi(), sigma_to_outlier(), dpsi_max();
  float outlier, u, v, dscale;

  outlier = sigma_to_outlier(sigma);

  x_cent = (int)floor(nx/2.0);
  y_cent = (int)floor(ny/2.0);

  dscale = dpsi_max(sigma);
  
  top = 0.0; scale=0.0;
  for(i=0;i<ny;i++){
    for(j=0;j<nx;j++){
      x = j - x_cent;
      y = i - y_cent;
      index = (i*nx) + j;
      if (omask[index] >= 0.5) {
	u = *a0 + (*a1*x) + (*a2*y) + (*a6*x*x) + (*a7*x*y);
	v = *a3 + (*a4*x) + (*a5*y) + (*a6*x*y) + (*a7*y*y);
	err = (Ix[index]*u) + (Iy[index]*v) + It[index];
	gr = Ix[index];
	top += gr * psi(err, sigma);
	scale += (gr*gr)*dscale; 
      }
    }
  }
  if (scale == 0.0) 
    delta = 0.0;
  else
    delta = (omega * top/scale);
  max_delta = max(abs(delta), max_delta);
  *a0 -=  delta;

  top = 0.0; scale=0.0;
  for(i=0;i<ny;i++){
    for(j=0;j<nx;j++){
      x = j - x_cent;
      y = i - y_cent;
      index = (i*nx) + j;
      if (omask[index] >= 0.5) {
	u = *a0 + (*a1*x) + (*a2*y) + (*a6*x*x) + (*a7*x*y);
	v = *a3 + (*a4*x) + (*a5*y) + (*a6*x*y) + (*a7*y*y);
	err = (Ix[index]*u) + (Iy[index]*v) + It[index];
	gr = Iy[index];
	top += gr * psi(err, sigma);
	scale += (gr*gr)*dscale; 
      }
    }
  }
  if (scale == 0.0) 
    delta = 0.0;
  else
    delta = (omega * top/scale);

  max_delta = max(abs(delta), max_delta);
  *a3 -=  delta;

  if ((affine != 0) && ((nx>=MIN_AFFINE_SIZE) || (ny>=MIN_AFFINE_SIZE))) {
    top = 0.0; scale=0.0;
    for(i=0;i<ny;i++){
      for(j=0;j<nx;j++){
	x = j - x_cent;
	y = i - y_cent;
	index = (i*nx) + j;
	if (omask[index] >= 0.5) {
	  u = *a0 + (*a1*x) + (*a2*y) + (*a6*x*x) + (*a7*x*y);
	  v = *a3 + (*a4*x) + (*a5*y) + (*a6*x*y) + (*a7*y*y);
	  err = (Ix[index]*u) + (Iy[index]*v) + It[index];
	  gr = Ix[index];
	  top += x * gr * psi(err, sigma);
	  scale += (x*x*gr*gr)*dscale; 
	}
      }
    }
    if (scale == 0.0) 
    delta = 0.0;
  else
    delta = (omega * top/scale);

    max_delta = max(abs(delta), max_delta);
    *a1 -=  delta;
    
    top = 0.0; scale=0.0;
    for(i=0;i<ny;i++){
      for(j=0;j<nx;j++){
	x = j - x_cent;
	y = i - y_cent;
	index = (i*nx) + j;
	if (omask[index] >= 0.5) {
	  u = *a0 + (*a1*x) + (*a2*y) + (*a6*x*x) + (*a7*x*y);
	  v = *a3 + (*a4*x) + (*a5*y) + (*a6*x*y) + (*a7*y*y);
	  err = (Ix[index]*u) + (Iy[index]*v) + It[index];
	  gr = Ix[index];
	  top += y * gr * psi(err, sigma);
	  scale += (y*y*gr*gr)*dscale; 
	}
      }
    }
    if (scale == 0.0) 
    delta = 0.0;
  else
    delta = (omega * top/scale);

    max_delta = max(abs(delta), max_delta);
    *a2 -=  delta;
    
    top = 0.0; scale=0.0;
    for(i=0;i<ny;i++){
      for(j=0;j<nx;j++){
	x = j - x_cent;
	y = i - y_cent;
	index = (i*nx) + j;
	if (omask[index] >= 0.5) {
	  u = *a0 + (*a1*x) + (*a2*y) + (*a6*x*x) + (*a7*x*y);
	  v = *a3 + (*a4*x) + (*a5*y) + (*a6*x*y) + (*a7*y*y);
	  err = (Ix[index]*u) + (Iy[index]*v) + It[index];
	  gr = Iy[index];
	  top += x * gr * psi(err, sigma);
	  scale += (x*x*gr*gr)*dscale; 
	}
      }
    }
    if (scale == 0.0) 
      delta = 0.0;
    else
      delta = (omega * top/scale);
    max_delta = max(abs(delta), max_delta);
    *a4 -=  delta;
    
    top = 0.0; scale=0.0;
    for(i=0;i<ny;i++){
      for(j=0;j<nx;j++){
	x = j - x_cent;
	y = i - y_cent;
	index = (i*nx) + j;
	if (omask[index] >= 0.5) {
	  u = *a0 + (*a1*x) + (*a2*y) + (*a6*x*x) + (*a7*x*y);
	  v = *a3 + (*a4*x) + (*a5*y) + (*a6*x*y) + (*a7*y*y);
	  err = (Ix[index]*u) + (Iy[index]*v) + It[index];
	  if (abs(err)>outlier)
	    num_outliers++;
	  gr = Iy[index];
	  top += y * gr * psi(err, sigma);
	  scale += (y*y*gr*gr)*dscale; 
	}
      }
    }
    if (scale == 0.0) 
      delta = 0.0;
    else
      delta = (omega * top/scale);
    max_delta = max(abs(delta), max_delta);
    *a5 -=  delta;
  }

  if ((planar != 0)  && ((nx>=MIN_PLANAR_SIZE) || (ny>=MIN_PLANAR_SIZE))) {
    top = 0.0; scale=0.0;
    for(i=0;i<ny;i++){
      for(j=0;j<nx;j++){
	x = j - x_cent;
	y = i - y_cent;
	index = (i*nx) + j;
	if (omask[index] >= 0.5) {
	  u = *a0 + (*a1*x) + (*a2*y) + (*a6*x*x) + (*a7*x*y);
	  v = *a3 + (*a4*x) + (*a5*y) + (*a6*x*y) + (*a7*y*y);
	  err = (Ix[index]*u) + (Iy[index]*v) + It[index];
	  if (abs(err)>outlier)
	    num_outliers++;
	  gr = ((x * x * Ix[index])+(x*y*Iy[index]));
	  top += gr * psi(err, sigma);
	  scale += (gr*gr)*dscale; 
	}
      }
    }
    if (scale == 0.0) 
      delta = 0.0;
    else
      delta = (omega * top/scale);
    max_delta = max(abs(delta), max_delta);
    *a6 -=  delta;

    top = 0.0; scale=0.0;
    for(i=0;i<ny;i++){
      for(j=0;j<nx;j++){
	x = j - x_cent;
	y = i - y_cent;
	index = (i*nx) + j;
	if (omask[index] >= 0.5) {
	  u = *a0 + (*a1*x) + (*a2*y) + (*a6*x*x) + (*a7*x*y);
	  v = *a3 + (*a4*x) + (*a5*y) + (*a6*x*y) + (*a7*y*y);
	  err = (Ix[index]*u) + (Iy[index]*v) + It[index];
	  if (abs(err)>outlier)
	    num_outliers++;
	  gr = ((x * y * Ix[index])+(y*y*Iy[index]));
	  top += gr * psi(err, sigma);
	  scale += (gr*gr)*dscale; 
	}
      }
    }
    if (scale == 0.0) 
      delta = 0.0;
    else
      delta = (omega * top/scale);
    max_delta = max(abs(delta), max_delta);
    *a7 -=  delta;
  }

  /* fprintf(stderr, "affine outliers %d\n", num_outliers); */
  return(max_delta);
}


float regress_affine(im1, im2, a0, a1, a2, a3, a4, a5, a6, a7, Ix, Iy, It, nx, ny, 
		  iters, sigma, sigma_init, omega, omask, rate, affine, planar)
     float *im1, *im2, *a0, *a1, *a2, *a3, *a4, *a5, *a6, *a7, *Ix, *Iy, *It, 
       sigma, sigma_init, omega, *omask;
     int iters,  nx, ny, affine, planar;
     float rate;
{
  int i;
  float max_delta, s;

  s = sigma_init;
  fprintf(stderr, "Affine;  a0: %f, a3: %f\n ", *a0, *a3);
    for(i=0;i<iters;i++){
      max_delta =
	affine_regress_1_iter(Ix, Iy, It, a0, a1, a2, a3, a4, a5, a6, a7, nx, ny, 
			      s, omega, omask, affine, planar);
      s = max((s * rate), sigma);
      fprintf(stderr, ".");
/*
      fprintf(stderr, "Iter %d, a1: %f a2: %f a4: %f a5: %f \n", 
	      i, *a1, *a2, *a4, *a5);
*/
      if (max_delta <= AFFINE_THRESHOLD) {
	fprintf(stderr, "STOP\n");
	i = iters;
      }
    }
  
  fprintf(stderr, "\n min sigma %f\n\n", s);
  return(s);
}  


outliers(Ix, Iy, It, a0, a1, a2, a3, a4, a5, a6, a7, nx, ny, sigma, outlier_im)

     float *Ix, *Iy, *It, a0, a1, a2, a3, a4, a5, a6, a7, sigma, *outlier_im;
     int nx, ny;
{
  int index, i, j, x_cent, y_cent, jpos, ipos;
  float err, outlier, u, v, sigma_to_outlier();

  outlier = sigma_to_outlier(sigma);

  x_cent = (int)floor(nx/2.0);
  y_cent = (int)floor(ny/2.0);


  for(i=0;i<ny;i++){
    for(j=0;j<nx;j++){
      outlier_im[(i*nx)+j] = 0.0;
    }
  }

  for(i=0;i<ny;i++){
    for(j=0;j<nx;j++){
      index = (i*nx) + j;
      jpos = j-x_cent;
      ipos = i-y_cent;
      u = a0 + (jpos * a1) + (ipos * a2) + (jpos*jpos*a6) + (ipos*jpos*a7);
      v = a3 + (jpos * a4) + (ipos * a5) + (ipos*jpos*a6) + (ipos*ipos*a7);
      err = Ix[index]*u + Iy[index]*v + It[index];
      if (abs(err)>outlier) {
	outlier_im[index] = 1.0;
      }
    }
  }

}

good_points(Ix, Iy, It, a0, a1, a2, a3, a4, a5, a6, a7, nx, ny, sigma, outlier_im)

     float *Ix, *Iy, *It, a0, a1, a2, a3, a4, a5, a6, a7, sigma, *outlier_im;
     int nx, ny;
{
  int index, i, j, x_cent, y_cent, jpos, ipos;
  float err, outlier, u, v, sigma_to_outlier();

  outlier = sigma_to_outlier(sigma);

  x_cent = (int)floor(nx/2.0);
  y_cent = (int)floor(ny/2.0);

/*
  for(i=0;i<ny;i++){
    for(j=0;j<nx;j++){
      outlier_im[(i*nx)+j] = 1.0;
    }
  }
*/
  for(i=0;i<ny;i++){
    for(j=0;j<nx;j++){
      index = (i*nx) + j;
      jpos = j-x_cent;
      ipos = i-y_cent;
      u = a0 + (jpos * a1) + (ipos * a2) + (jpos*jpos*a6) + (ipos*jpos*a7);
      v = a3 + (jpos * a4) + (ipos * a5) + (ipos*jpos*a6) + (ipos*ipos*a7);
      err = Ix[index]*u + Iy[index]*v + It[index];
      if (outlier_im[index] == 0.0)  /* from warping */
	outlier_im[index] = 1.0;
      else if (abs(err)>outlier) {
	outlier_im[index] = 0.0;
      }
    }
  }

}



robust_to_weights(Ix, Iy, It, a0, a1, a2, a3, a4, a5, a6, a7, nx, ny, sigma, weight_im)

     float *Ix, *Iy, *It, a0, a1, a2, a3, a4, a5, a6, a7, sigma, *weight_im;
     int nx, ny;
{
  int index, i, j, x_cent, y_cent, jpos, ipos;
  float err, u, v, psi(), outlier, max_weight, sigma_to_outlier();

  x_cent = (int)floor((float)nx/2.0);
  y_cent = (int)floor((float)ny/2.0);

  outlier = sigma_to_outlier(sigma);
  max_weight = psi(outlier, sigma)/(outlier);
  fprintf(stderr, "sigma %f, outlier %f, max_weight %f\n", sigma, outlier, max_weight);

  for(i=0;i<ny;i++){
    for(j=0;j<nx;j++){
      index = (i*nx) + j;
      jpos = j-x_cent;
      ipos = i-y_cent;
      u = a0 + (jpos * a1) + (ipos * a2) + (jpos*jpos*a6) + (ipos*jpos*a7);
      v = a3 + (jpos * a4) + (ipos * a5) + (ipos*jpos*a6) + (ipos*ipos*a7);
      err = Ix[index]*u + Iy[index]*v + It[index];
      if (err==0.0) {
	weight_im[index] = max_weight;
      }
      else {
	weight_im[index] = psi(err, sigma)/(2.0*err);
      }
    }
  }
}
